#include "edge.h"
#include "tile.h"

Edge::Edge(Tile *pParent): m_pParent(pParent), m_pNeighbor(0), m_type(FARM)
{
	for(int i=0;i<3;i++)
		m_features[i] = 0;
}

Edge::~Edge()
{

}

void Edge::Display()
{
	for(int i=0;i<3;i++)
		cout << m_features[i]->m_number << " ";
	cout << endl;
}

bool Edge::HasFeature(Feature *pFeature)
{
	for(int i=0;i<3;i++)
		if(m_features[i] == pFeature)
			return true;

	return false;
}

void Edge::SetFeatures(string data, vector <Feature *>& features)
{
	for(int i=0;i<3;i++) {
		// (char one-based -> int zero-based)
		m_features[i] = features[data[i] - '1'];
		if(m_features[i])
			m_features[i]->IncreaseBorderCount();
	}
}

// AttachTo
// . Attaches us to the given edge.
// . Returns a list of Features to be deleted.
// . Updates the tile counts for each of our features.
vector <Feature *> Edge::AttachTo(Edge *pEdge)
{
	vector <Feature *> ret;

	// Make us neighbors
	SetNeighbor(pEdge);
	pEdge->SetNeighbor(this);
	
	// Go through the three border locations and:
	// . Merge our feature onto their feature
	// . Replace our feature by their feature
	// . Add the to-be-deleted feature to the list
	for(int i=0;i<3;i++) {
		// Remember, opposite edges are indexed backwards
		// since everything is clockwise, so we need to
		// grab our i-th feature and his (2-i)-th feature.
		Feature *pOut = GetFeature(i),
			*pIn = pEdge->GetFeature(2 - i);

		pOut->DropBorderCount();
		pIn->DropBorderCount();

		// Already the same Feature? Just go on.
		if(pOut == pIn)
			continue;
					
		pIn->Merge(pOut);
		Replace(pOut, pIn);

		ret.push_back(pOut);
	}

	// Update the tile counts for the features that have been changed.
	for(int i=0;i<3;i++) {
		map <Tile *, bool> visited;

		m_features[i]->ZeroTileCount();
		m_pParent->CheckTileCount(m_features[i], visited);
	}

	return ret;
}

// Replace
// . Replaces all instances of 'pOut' with 'pIn'.
// . If 'doParent' is true, then recurse on to our parent.
// . If we found an instance and we have a neighbor, then
//   recurse on to him.
void Edge::Replace(Feature *pOut, Feature *pIn, bool doParent)
{
	bool found = false;
	for(int i=0;i<3;i++) {
		if(m_features[i] == pOut) {
			m_features[i] = pIn;
			found = true;
		}
	}

	if(doParent)
		m_pParent->Replace(pOut, pIn);

	if(found && m_pNeighbor)
		m_pNeighbor->Replace(pOut, pIn);
}
