#include "sdlwindow.h"
#include "sdlinput.h"

Window::Window()
{
	m_pInput = new Input(this);
}

Window::~Window()
{
	Quit();
	delete m_pInput;
}

// Run
// . This is the main function that should be called
bool Window::Run(int width, int height, int bpp)
{
	if(!InitSDL(width, height, bpp))
		return false;

	Initialize();
	DoLoop();
	Quit();
	
	return true;
}

// InitSDL
// . Initializes an SDL window that runs OpenGL
bool Window::InitSDL(int width, int height, int bpp)
{
	if(SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Video initialization failed: %s\n", 	SDL_GetError());
		Quit();
		return false;
	}

	const SDL_VideoInfo *pVideoInfo = SDL_GetVideoInfo();

	if(!pVideoInfo) {
		fprintf(stderr, "Video query failed: %s\n", SDL_GetError());
		Quit();
		return false;
	}

	// These flags are:
	// OpenGL, double buffering, store palette in hardware, window resizing
	m_videoFlags = SDL_OPENGL | SDL_GL_DOUBLEBUFFER | SDL_HWPALETTE | SDL_RESIZABLE;

	// Can surfaces be stored in memory?
	if(pVideoInfo->hw_available)
		m_videoFlags |= SDL_HWSURFACE;
	else
		m_videoFlags |= SDL_SWSURFACE;

	// Can we do hardware blits?
	if(pVideoInfo->blit_hw)
		m_videoFlags |= SDL_HWACCEL;

	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

	m_pSurface = SDL_SetVideoMode(width, height, bpp, m_videoFlags);

	if(!m_pSurface) {
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
		Quit();
		return false;
	}

	if(!InitGL()) {
		fprintf(stderr, "Could not initialize OpenGL.\n");
		Quit();
		return false;
	}

//	Resize(width, height, true, 1000);
	m_hasFocus = true;
	m_isActive = true;

	return true;
}

// InitGL
// . Sets up basic OpenGL settings
bool Window::InitGL()
{
	glShadeModel(GL_SMOOTH);
	glClearColor(0, 0, 0, 0);
	glClearDepth(1);
//	glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//	glDepthFunc(GL_LEQUAL);

	return true;
}

// Quit
// . Destroys the SDL window
void Window::Quit()
{
	SDL_Quit();
	m_hasFocus = false;
	m_isActive = false;
}

// ToggleFullScreen
// . Yeah.
void Window::ToggleFullScreen()
{
	SDL_WM_ToggleFullScreen(m_pSurface);
}

// DoLoop
// . This is the main loop. It reads messages,
//   calls Draw() and Update(...) once per frame.
void Window::DoLoop()
{
	int time = 0, frames = 0;

	m_ticks = SDL_GetTicks();
	while(HandleEvents() && m_isActive) {
		int ticks = SDL_GetTicks(), dTicks = ticks - m_ticks;
		m_ticks = ticks;

		if(m_hasFocus) {
			Update(.001 * dTicks);

			Draw();
			SDL_GL_SwapBuffers();
			SDL_Delay(1);
			
			time += dTicks;
			frames++;
			if(time >= 1000) {
				time -= 1000;
				m_FPS = frames;
				frames = 0;
			}
		}
	}
}

// HandleEvents
// . This reads SDL events from the queue and
//   handles them. Input is passed on to the
//   Input class.
bool Window::HandleEvents()
{
	SDL_Event event;

	while(SDL_PollEvent(&event)) {
		switch(event.type) {
			case SDL_ACTIVEEVENT:
      		if(event.active.state == SDL_APPINPUTFOCUS) {
					if(event.active.gain == 0)
						m_hasFocus = false;
					else
						m_hasFocus = true;
				}
				break;

			case SDL_VIDEORESIZE:
				m_pSurface = SDL_SetVideoMode(event.resize.w, event.resize.h, 16, m_videoFlags);
				if(!m_pSurface) {
					fprintf(stderr, "Could not get a surface after resize: %s\n", SDL_GetError());
					Quit();
					return false;
				}

				break;


			case SDL_KEYDOWN:
				m_pInput->KeyDown(event.key);
				break;

			case SDL_KEYUP:
				m_pInput->KeyUp(event.key);
				break;

			case SDL_MOUSEMOTION:
				m_pInput->MoveMouse(event.motion);
				break;
				
			case SDL_MOUSEBUTTONDOWN:
				m_pInput->ButtonDown(event.button);
				break;

			case SDL_MOUSEBUTTONUP:
				m_pInput->ButtonUp(event.button);
				break;

			case SDL_QUIT:
				return false;
				break;

			default:
				break;
		}
	}

	return true;
}

/////////////////////////////////////////////////////////////
// Create(*)Window(*)
// . These functions create a subwindow that runs OpenGL
//   inside the main window.
// . The inputs to the ...f functions are [0, 1] x [0, 1]
//   coordinates of the window.
// . The inputs to the ...i functions are [0, width] x [0, height]
//   coordinates of the window.

bool Window::CreatePerspectiveWindowf(double x, double y, double X, double Y, double far, double perspective)
{
	if(!m_pSurface)
		return false;

	int width = m_pSurface->w, height = m_pSurface->h;

	return CreatePerspectiveWindowi((int) (x * width), (int) (y * height), (int) ((X - x) * width), (int) ((Y - y) * height), far, perspective);
}

bool Window::CreateOrthoWindowf(double x, double y, double X, double Y)
{
	if(!m_pSurface)
		return false;

	int width = m_pSurface->w, height = m_pSurface->h;

	return CreateOrthoWindowi((int) (x * width), (int) (y * height), (int) ((X - x) * width), (int) ((Y - y) * height));
}

bool Window::CreatePerspectiveWindowi(int x, int y, int width, int height, double far, double perspective)
{
	if(height == 0)
		height = 1;

	double ratio = (double) width / height;

	glViewport(x, y, x + width, y + height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(perspective, ratio, .1, far);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	return true;
}

bool Window::CreateOrthoWindowi(int x, int y, int width, int height)
{
	if(height == 0)
		height = 1;

	glViewport(x, y, x + width, y + height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glOrtho(0, width, height, 0, -1, 1);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	return true;
}
