#include "tile.h"
#include "farm.h"
#include "road.h"
#include "city.h"
#include "cloister.h"
#include "river.h"

#include <iostream>
#include <fstream>
using namespace std;

Tile::Tile()
{
	for(int i=0;i<4;i++)
		m_edges[i] = new Edge(this);

	m_pCloister = 0;
	m_orientation = NORTH;
	m_pMeeple = 0;
}

Tile::~Tile()
{
	for(int i=0;i<4;i++)
		delete m_edges[i];
}

bool Tile::IsRiver()
{
	for(int i=0;i<4;i++)
		if(m_edges[i]->GetType() == RIVER)
			return true;

	return false;
}

bool Tile::HasFeature(Feature *pFeature)
{
	for(int i=0;i<4;i++)
		if(m_edges[i]->HasFeature(pFeature))
			return true;

	return false;
}

Feature *Tile::GetFeatureAt(int i, int j)
{
	Point q = RotateFrom(Point(i, j), m_orientation, 2);
	if(m_entryPoints.count(q))
		return m_entryPoints[q];
	return 0;
}

void Tile::Draw(double x, double y, double sX, double sY, bool forceTo, bool entryPoints)
{
	int	X = (forceTo ? 0 : m_p.m_x),
			Y = (forceTo ? 0 : m_p.m_y);

	glTranslatef(x + X * sX, y + Y * sY, 0);
	
	glTranslatef(.5 * sX, .5 * sY, 0);
	glRotatef(GetRotation(m_orientation), 0, 0, 1);
	glTranslatef(-.5 * sX, -.5 * sY, 0);

	glColor3f(1, 1, 1);
	m_image.Bind();

	glBegin(GL_QUADS);
		glTexCoord2f(0, 0); glVertex2f(0, 0);
		glTexCoord2f(1, 0); glVertex2f(sX, 0);
		glTexCoord2f(1, 1); glVertex2f(sX, sY);
		glTexCoord2f(0, 1); glVertex2f(0, sY);
	glEnd();

	m_image.UnBind();

	glTranslatef(.5 * sX, .5 * sY, 0);
	glRotatef(-GetRotation(m_orientation), 0, 0, 1);
	glTranslatef(-.5 * sX, -.5 * sY, 0);

	if(m_pMeeple)
		m_pMeeple->Draw((double) sX / 3);

	if(entryPoints) {
		glColor4f(1, 0, 0, .5);
		glBegin(GL_QUADS);
			double tX = (double) sX / 3, tY = (double) sY / 3;
			for(int i=0;i<3;i++) {
				for(int j=0;j<3;j++) {
					if(CanAddMeeple(i, j)) {
						glVertex2f(i * tX,		j * tY);
						glVertex2f((i + 1) * tX,	j * tY);
						glVertex2f((i + 1) * tX,	(j + 1) * tY);
						glVertex2f(i * tX,		(j + 1) * tY);
					}
				}
			}
		glEnd();
	}

	glTranslatef(-x - X * sX, -y - Y * sY, 0);
}

void Tile::Display()
{

	cout << "Tile at (" << m_p.m_x << ", " << m_p.m_y << "):\n";

	for(int i=0;i<4;i++)
		GetEdge((DIR) i)->Display();
}

// Create
// . Creates the tile with the given specifications.
// . Remember, it's up to the caller of this function
//   to destroy the Features that it creates!
vector <Feature *> Tile::Create(string name)
{
	vector <Feature *> ret;
	Feature *pFeature = 0;

	m_image.LoadBitmap(name + ".bmp", LINEAR);

	ifstream fin((name + ".tile").c_str());

	// Grab the edge types
	string str;
	for(int i=0;i<4;i++) {
		fin >> str;
		m_edges[i]->SetType(str);
	}

	// Grab all of the features
	while(1) {
		fin >> str;
		if(str == "*")
			break;

		switch(str[0]) {
			case 'f': pFeature = new Farm; break;
			case 'r': pFeature = new Road; break;
			case 'c': pFeature = new City; break;
			case 'k': pFeature = new Cloister; break;
			case 'v': pFeature = new River; break;
			default: pFeature = 0;
		}
		
		if(pFeature)
			ret.push_back(pFeature);

		if(str[0] == 'k')
			m_pCloister = dynamic_cast <Cloister *> (pFeature);

		if(str[0] != 'v') {
			Point p(str[1] - '1', str[2] - '1');
			m_entryPoints[p] = pFeature;
		}
	}

	// Read the features for each edge
	for(int i=0;i<4;i++) {
		fin >> str;
		m_edges[i]->SetFeatures(str, ret);
	}

	fin.close();
	
	return ret;
}

bool Tile::CanAttach(Tile *pTile, DIR direction, bool& tentative)
{
	Edge *pUs = GetEdge(direction), *pThem = pTile->GetEdge(Opposite(direction));
	if(pUs->GetType() != pThem->GetType()) {
		cout << "Tile does not mesh at " << Dir[direction] << "ern edge: " <<
				EdgeType[pUs->GetType()] << " != " << EdgeType[pThem->GetType()] << endl;
		return false;
	}

	if(IsRiver() && pTile->IsRiver()) {
		if(GetEdge(direction)->GetType() == RIVER) {
			tentative = true;

			for(int i=0;i<4;i++) {
				if(i == direction || i == Opposite(direction))
					continue;

				if(GetEdge((DIR) i)->GetType() == RIVER && pTile->GetEdge((DIR) i)->GetType() == RIVER) {
					cout << "No U-turns with river pieces!\n";
					return false;
				}
			}
		}
	} else
		tentative = true;

	return true;
}

// Attach
// . Attaches this tile to the specified tile, in
//   the given direction.
// . Returns the Features that ought to be deleted.
// . The Features are deleted from THIS tile, not
//   from the neighbor's.
vector <Feature *> Tile::Attach(Tile *pTile, DIR direction)
{
	Edge	*pUs = GetEdge(direction),
		*pThem = pTile->GetEdge(Opposite(direction));

	return pUs->AttachTo(pThem);
}

// Replace
// . Replaces every instance of 'pOut' with 'pIn'.
// . If any instance of 'pOut' was found, then it
//   recurses through our neighbors.
void Tile::Replace(Feature *pOut, Feature *pIn)
{
	for(int i=0;i<4;i++)
		m_edges[i]->Replace(pOut, pIn, false);

	for(FeatureIt i=m_entryPoints.begin();i!=m_entryPoints.end();i++)
		if(i->second == pOut)
			i->second = pIn;
}

void Tile::AddNeighbor()
{
	if(m_pCloister)
		m_pCloister->DropBorderCount();
}

// TileCount
// . Recurses through the tiles, tallying up how many
//   tiles the given Feature is found on.
void Tile::CheckTileCount(Feature *pFeature, map <Tile *, bool>& visited)
{
	if(visited.count(this))
		return;

	visited[this] = true;

	if(!HasFeature(pFeature))
		return;

	pFeature->IncreaseTileCount();
	for(int i=0;i<4;i++) {
		Tile *pNeighbor = m_edges[i]->GetNeighborTile();
		if(pNeighbor && !visited.count(pNeighbor))
			pNeighbor->CheckTileCount(pFeature, visited);
	}
}

bool Tile::CanAddMeeple(int i, int j)
{
	if(m_pMeeple)
		return false;

	Feature *pFeature = GetFeatureAt(i, j);

	if(!pFeature)
		return false;

	if(pFeature->NumOfMeeples() > 0)
		return false;

	return true;
}

bool Tile::AddMeeple(Meeple *pMeeple, int i, int j)
{
	if(!CanAddMeeple(i, j))
		return false;

	m_pMeeple = pMeeple;
	m_pMeeple->Place(GetFeatureAt(i, j), this, Point(i, j));

	return true;
}

void Tile::ReturnMeeple()
{
	m_pMeeple = 0;
}
