///////////////////////////////////////////////////////
// Tile
//
// The Tile class contains all information contained in
// a single tile:
//
// . Each edge is labelled as FARM, ROAD, or CITY
// . Each edge has a pointer to its next-door neighbor.
// . Each edge has a 1x3 grid, each containing a pointer to a Feature.
// . When a Tile is created, a Feature is created for each
//   *distinct* location on the tile, and then each square
//   in the grid points to the correct Feature.
//
// . For example, suppose a tile has a single horizontal road.
//   . The edges, beginning at the top going clockwise, would
//     be labelled FARM, ROAD, FARM, ROAD.
//   . Three distinct features would be created:
//     A) Farm
//     B) Road
//     C) Farm
//   . The grid would look like: AAA
//                               ABC
//                               CCC
//                               CBA
//     where A, B, and C are pointers to the above features.
//
//  . Important Note:
//    . A tile is not created by the constructor, but instead
//      by the function Create(...). This function returns a
//      vector containing the Features that it created.
//    . Destruction of the Features is *not* done by the Tile,
//      but by whichever class called Create(...).

#pragma once

#include "feature.h"
#include "edge.h"
#include "point.h"
#include "image.h"
#include "cloister.h"
#include "meeple.h"
#include "player.h"

#include <vector>
#include <map>
#include <string>
using namespace std;

typedef map <Point, Feature *>::iterator FeatureIt;

class Tile
{
public:
	Tile();
	~Tile();

	void Display();

	bool IsRiver();
	bool HasFeature(Feature *pFeature);
	Edge *GetEdge(DIR direction) const { return m_edges[LocalDir(direction)]; }
	Point GetPoint() const { return m_p; }
	Feature *GetFeatureAt(int i, int j);

	void SetPoint(int x, int y) { m_p.Set(x, y); }
	void SetPoint(Point p) { m_p = p; }
	void SetOrientation(DIR orientation) { m_orientation = orientation; }
	void RotateCW() { m_orientation = TurnCW(m_orientation); }
	void RotateCCW() { m_orientation = TurnCCW(m_orientation); }
	DIR LocalDir(DIR direction) const { return RotateBy(direction, m_orientation); }

	void Draw(double x, double y, double sX, double sY, bool forceTo = false, bool entryPoints = false);

	vector <Feature *> Create(int i);
	vector <Feature *> Create(string name);
	bool CanAttach(Tile *pTile, DIR direction, bool& tentative);
	vector <Feature *> Attach(Tile *pTile, DIR direction);

	void Replace(Feature *pOut, Feature *pIn);
	void CheckTileCount(Feature *pFeature, map <Tile *, bool>& visited);

	void AddNeighbor();
	bool CanAddMeeple(int i, int j);
	bool AddMeeple(Meeple *pMeeple, int i, int j);
	void ReturnMeeple();

private:
	Point m_p;
	DIR m_orientation;
	Image m_image;

	Edge *m_edges[4];
	Cloister *m_pCloister;
	map <Point, Feature *> m_entryPoints;

	Meeple *m_pMeeple;
};
